class Party(object):
    """Trading party.

    Represents either a consumer or business in a given country acting as a
    party to a transaction.

    :ivar country_code:
        Party's legal or effective country of registration or residence as an
        ISO 3166-1 alpha-2 country code.
    :type country_code: str
    :ivar is_business: Whether the part is a legal business entity.
    :type is_business: bool
    """

    def __init__(self, country_code, is_business):
        """Initialize a trading party.

        :param country_code:
            Party's legal or effective country of registration or residence as
            an ISO 3166-1 alpha-2 country code.
        :type country_code: str
        :param is_business: Whether the part is a legal business entity.
        :type is_business: bool
        """

        self.country_code = country_code
        self.is_business = is_business

    def __repr__(self):
        return '<pyvat.Party: country code = %s, is business = %r>' % (
            self.country_code, self.is_business,
        )



EU_COUNTRY_CODES = set([
    'AT',  # Austria.
    'BE',  # Belgium.
    'BG',  # Bulgaria.
    'CY',  # Cyprus.
    'CZ',  # Czech Republic.
    'DE',  # Germany.
    'DK',  # Denmark.
    'EE',  # Estonia.
    'ES',  # Spain.
    'FI',  # Finland.
    'FR',  # France.
    'EL', 'GR',  # Greece.
    'HR',  # Croatia.
    'HU',  # Hungary.
    'IE',  # Ireland.
    'IT',  # Italy.
    'LT',  # Lithuania.
    'LU',  # Luxembourg.
    'LV',  # Latvia.
    'MT',  # Malta.
    'NL',  # Netherlands.
    'PL',  # Poland.
    'PT',  # Portugal.
    'RO',  # Romania.
    'SE',  # Sweden.
    'SI',  # Slovenia.
    'SK',  # Slovakia.
])
"""EU country codes.

Represented by ISO 3166-1 alpha-2 country codes.
"""



import datetime
from decimal import Decimal
from .countries import EU_COUNTRY_CODES
from .item_type import ItemType
from .vat_charge import VatCharge, VatChargeAction
from .utils import ensure_decimal

JANUARY_1_2015 = datetime.date(2015, 1, 1)


class VatRules(object):
    """Base VAT rules for a country.
    """

    def get_vat_rate(self, item_type):
        """Get the VAT rate for an item type.

        :param item_type: Item type.
        :type item_type: ItemType
        :returns: the VAT rate in percent.
        :rtype: decimal.Decimal
        """

        raise NotImplementedError()

    def get_sale_to_country_vat_charge(self,
                                       date,
                                       item_type,
                                       buyer,
                                       seller):
        """Get the VAT charge for selling an item to a buyer in the country.

        :param date: Sale date.
        :type date: datetime.date
        :param item_type: Type of the item being sold.
        :type item_type: ItemType
        :param buyer: Buyer.
        :type buyer: Party
        :param seller: Seller.
        :type seller: Party
        :rtype: VatCharge
        :raises NotImplementedError:
            if no explicit rules for selling to the given country from the
            given country exist. VAT charge determining falls back to testing
            rules for the country in which the seller resides if this is
            raised.
        """

        raise NotImplementedError()

    def get_sale_from_country_vat_charge(self,
                                         date,
                                         item_type,
                                         buyer,
                                         seller):
        """Get the VAT charge for selling an item as a seller in the country.

        :param date: Sale date.
        :type date: datetime.date
        :param item_type: Type of the item being sold.
        :type item_type: ItemType
        :param buyer: Buyer.
        :type buyer: Party
        :param seller: Seller.
        :type seller: Party
        :rtype: VatCharge
        :raises NotImplementedError:
            if no rules for selling from the given country to the given country
            exist.
        """

        raise NotImplementedError()


class EuVatRulesMixin(object):
    """Mixin for VAT rules in EU countries.
    """

    def get_sale_to_country_vat_charge(self,
                                       date,
                                       item_type,
                                       buyer,
                                       seller):
        # We only support business sellers at this time.
        if not seller.is_business:
            raise NotImplementedError(
                'non-business sellers are currently not supported'
            )

        # If the seller resides in the same country as the buyer, we charge
        # VAT regardless of whether the buyer is a business or not. Similarly,
        # if the buyer is a consumer, we must charge VAT in the buyer's country
        # of residence.
        if seller.country_code == buyer.country_code or \
                (not buyer.is_business and date >= JANUARY_1_2015):
            return VatCharge(VatChargeAction.charge,
                             buyer.country_code,
                             self.get_vat_rate(item_type))

        # EU consumers are charged VAT in the seller's country prior to January
        # 1st, 2015.
        if not buyer.is_business:
            # Fall back to the seller's VAT rules for this one.
            raise NotImplementedError()

        # EU businesses will never be charged VAT but must account for the VAT
        # by the reverse-charge mechanism.
        return VatCharge(VatChargeAction.reverse_charge,
                         buyer.country_code,
                         0)

    def get_sale_from_country_vat_charge(self,
                                         date,
                                         item_type,
                                         buyer,
                                         seller):
        # We only support business sellers at this time.
        if not seller.is_business:
            raise NotImplementedError(
                'non-business sellers are currently not supported'
            )

        # If the buyer resides outside the EU, we do not have to charge VAT.
        if buyer.country_code not in EU_COUNTRY_CODES:
            return VatCharge(VatChargeAction.no_charge, buyer.country_code, 0)

        # Both businesses and consumers are charged VAT in the seller's
        # country if both seller and buyer reside in the same country.
        if buyer.country_code == seller.country_code:
            return VatCharge(VatChargeAction.charge,
                             seller.country_code,
                             self.get_vat_rate(item_type))

        # Businesses in other EU countries are not charged VAT but are
        # responsible for accounting for the tax through the reverse-charge
        # mechanism.
        if buyer.is_business:
            return VatCharge(VatChargeAction.reverse_charge,
                             buyer.country_code,
                             0)

        # Consumers in other EU countries are charged VAT in their country of
        # residence after January 1st, 2015. Before this date, you charge VAT
        # in the country where the company is located.
        if date >= datetime.date(2015, 1, 1):
            buyer_rules = VAT_RULES[buyer.country_code]

            return VatCharge(VatChargeAction.charge,
                             buyer.country_code,
                             buyer_rules.get_vat_rate(item_type))
        else:
            return VatCharge(VatChargeAction.charge,
                             seller.country_code,
                             self.get_vat_rate(item_type))


class ConstantEuVatRateRules(EuVatRulesMixin):
    """VAT rules for a country with a constant VAT rate in the entiry country.
    """

    def __init__(self, vat_rate):
        self.vat_rate = ensure_decimal(vat_rate)

    def get_vat_rate(self, item_type):
        return self.vat_rate


class AtVatRules(EuVatRulesMixin):
    """VAT rules for Austria.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.prepaid_broadcasting_service:
            return Decimal(10)
        elif item_type == ItemType.ebook:
            return Decimal(10)
        return Decimal(20)


class CzVatRules(ConstantEuVatRateRules):
    """VAT rules for Czech Republic.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(10)

        return super(CzVatRules, self).get_vat_rate(item_type)


class BeVatRules(ConstantEuVatRateRules):
    """VAT rules for Belgium.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(6)

        return super(BeVatRules, self).get_vat_rate(item_type)


class IeVatRules(ConstantEuVatRateRules):
    """VAT rules for Ireland.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(9)

        return super(IeVatRules, self).get_vat_rate(item_type)


class FiVatRules(ConstantEuVatRateRules):
    """VAT rules for  Finland.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(10)

        return super(FiVatRules, self).get_vat_rate(item_type)


class NlVatRules(ConstantEuVatRateRules):
    """VAT rules for Netherlands.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(9)

        return super(NlVatRules, self).get_vat_rate(item_type)


class MtVatRules(ConstantEuVatRateRules):
    """VAT rules for Malta.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(5)

        return super(MtVatRules, self).get_vat_rate(item_type)


class SeVatRules(ConstantEuVatRateRules):
    """VAT rules for Sweden.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(6)

        return super(SeVatRules, self).get_vat_rate(item_type)


class HrVatRules(ConstantEuVatRateRules):
    """VAT rules for Croatia.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(5)

        return super(HrVatRules, self).get_vat_rate(item_type)


class PtVatRules(ConstantEuVatRateRules):
    """VAT rules for Portugal.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(6)

        return super(PtVatRules, self).get_vat_rate(item_type)


class FrVatRules(EuVatRulesMixin):
    """VAT rules for France.
    """

    def get_vat_rate(self, item_type):
        if item_type.is_broadcasting_service:
            return Decimal(10)
        if item_type == ItemType.ebook:
            return Decimal('5.5')
        if item_type == ItemType.enewspaper:
            return Decimal('2.1')
        return Decimal(20)


class ElVatRules(EuVatRulesMixin):
    """VAT rules for Greece.
    """

    def get_vat_rate(self, item_type):
        return Decimal(24)


class LuVatRules(EuVatRulesMixin):
    """VAT rules for Luxembourg.
    """

    def get_vat_rate(self, item_type):
        if item_type.is_broadcasting_service:
            return Decimal(3)
        elif item_type == ItemType.ebook:
            return Decimal(3)
        return Decimal(17)


class PlVatRules(EuVatRulesMixin):
    """VAT rules for Poland.
    """

    def get_vat_rate(self, item_type):
        if item_type.is_broadcasting_service:
            return Decimal(8)
        if item_type == ItemType.ebook:
            return Decimal(5)
        return Decimal(23)


class EsVatRules(EuVatRulesMixin):
    """VAT rules for Spain.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(4)
        return Decimal(21)


class DeVatRules(EuVatRulesMixin):
    """VAT rules for Germany.
    """

    def get_vat_rate(self, item_type):
        if item_type == ItemType.ebook:
            return Decimal(7)
        return Decimal(19)


# VAT rates are based on the report from January 1st, 2020
# http://ec.europa.eu/taxation_customs/sites/taxation/files/resources/documents/taxation/vat/how_vat_works/rates/vat_rates_en.pdf
VAT_RULES = {
    'AT': AtVatRules(),
    'BE': BeVatRules(21),
    'BG': ConstantEuVatRateRules(20),
    'CY': ConstantEuVatRateRules(19),
    'CZ': CzVatRules(21),
    'DE': DeVatRules(),
    'DK': ConstantEuVatRateRules(25),
    'EE': ConstantEuVatRateRules(20),
    'EL': ElVatRules(),
    'GR': ElVatRules(),  # Synonymous country code for Greece
    'ES': EsVatRules(),
    'FI': FiVatRules(24),
    'FR': FrVatRules(),
    'HR': HrVatRules(25),
    'HU': ConstantEuVatRateRules(27),
    'IE': IeVatRules(23),
    'IT': ConstantEuVatRateRules(22),
    'LT': ConstantEuVatRateRules(21),
    'LU': LuVatRules(),
    'LV': ConstantEuVatRateRules(21),
    'MT': MtVatRules(18),
    'NL': NlVatRules(21),
    'PL': PlVatRules(),
    'PT': PtVatRules(23),
    'RO': ConstantEuVatRateRules(19),
    'SE': SeVatRules(25),
    'SK': ConstantEuVatRateRules(20),
    'SI': ConstantEuVatRateRules(22),
}

"""VAT rules by country.

Maps an ISO 3316 alpha-2 country code to the VAT rules applicable in the given
country.
"""



import requests
import xml.dom.minidom

from requests import Timeout

from .result import VatNumberCheckResult
from .xml_utils import get_first_child_element, get_text, NodeNotFoundError
from .exceptions import ServerError


class Registry(object):
    """Abstract base registry.

    Defines an explicit interface for accessing arbitary registries.
    """

    def check_vat_number(self, vat_number, country_code):
        """Check if a VAT number is valid according to the registry.

        :param vat_number: VAT number without country code prefix.
        :param country_code: ISO 3166-1-alpha-2 country code.
        :returns: a :class:`VatNumberCheckResult` instance.
        """

        raise NotImplementedError()


class ViesRegistry(Registry):
    """VIES registry.

    Uses the European Commision's VIES registry for validating VAT numbers.
    """

    CHECK_VAT_SERVICE_URL = 'http://ec.europa.eu/taxation_customs/vies/' \
        'services/checkVatService'
    """URL for the VAT checking service.
    """

    DEFAULT_TIMEOUT = 8
    """Timeout for the requests."""

    def check_vat_number(self, vat_number, country_code):
        # Non-ISO code used for Greece.
        if country_code == 'GR':
            country_code = 'EL'

        # Request information about the VAT number.
        result = VatNumberCheckResult()

        request_data = (
                u'<?xml version="1.0" encoding="UTF-8"?><SOAP-ENV:Envelope'
                u' xmlns:ns0="urn:ec.europa.eu:taxud:vies:services:checkVa'
                u't:types" xmlns:ns1="http://schemas.xmlsoap.org/soap/enve'
                u'lope/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-insta'
                u'nce" xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/env'
                u'elope/"><SOAP-ENV:Header/><ns1:Body><ns0:checkVat><ns0:c'
                u'ountryCode>%s</ns0:countryCode><ns0:vatNumber>%s</ns0:va'
                u'tNumber></ns0:checkVat></ns1:Body></SOAP-ENV:Envelope>' %
                (country_code, vat_number)
        )

        result.log_lines += [
            u'> POST %s with payload of content type text/xml, charset UTF-8:',
            request_data,
        ]

        try:
            response = requests.post(
                self.CHECK_VAT_SERVICE_URL,
                data=request_data.encode('utf-8'),
                headers={
                    'Content-Type': 'text/xml; charset=utf-8',
                },
                timeout=self.DEFAULT_TIMEOUT
            )
        except Timeout as e:
            result.log_lines.append(u'< Request to EU VIEW registry timed out:'
                                    u' {}'.format(e))
            return result
        except Exception as exception:
            # Do not completely fail problematic requests.
            result.log_lines.append(u'< Request failed with exception: %r' %
                                    (exception))
            return result

        # Log response information.
        result.log_lines += [
            u'< Response with status %d of content type %s:' %
            (response.status_code, response.headers['Content-Type']),
            response.text,
            ]

        # Do not completely fail problematic requests.
        if response.status_code != 200 or \
                not response.headers['Content-Type'].startswith('text/xml'):
            result.log_lines.append(u'< Response is nondeterministic due to '
                                    u'invalid response status code or MIME '
                                    u'type')
            return result

        # Parse the DOM and validate as much as we can.
        #
        # We basically expect the result structure to be as follows,
        # where the address and name nodes might be omitted.
        #
        # <env:Envelope
	    #     xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">
        #         <env:Header/>
        #         <env:Body>
        #         <ns2:checkVatResponse
        #              xmlns:ns2="urn:ec.europa.eu:taxud:vies:services:checkVat:types">
        #              <ns2:countryCode>DE</ns2:countryCode>
        #              <ns2:vatNumber>812383453</ns2:vatNumber>
        #              <ns2:requestDate>2022-08-12+02:00</ns2:requestDate>
        #              <ns2:valid>true</ns2:valid>
        #              <ns2:name>---</ns2:name>
        #              <ns2:address>---</ns2:address>
        #         </ns2:checkVatResponse>
        #     </env:Body>
        # </env:Envelope>
        result_dom = xml.dom.minidom.parseString(response.text.encode('utf-8'))

        envelope_node = result_dom.documentElement
        if envelope_node.tagName != 'env:Envelope':
            raise ValueError(
                'expected response XML root element to be a SOAP envelope'
            )

        body_node = get_first_child_element(envelope_node, 'env:Body')

        # Check for server errors
        try:
            error_node = get_first_child_element(body_node, 'env:Fault')
            fault_strings = error_node.getElementsByTagName('faultstring')
            fault_code = fault_strings[0].firstChild.nodeValue
            raise ServerError(fault_code)
        except NodeNotFoundError:
            pass

        try:
            check_vat_response_node = get_first_child_element(
                body_node,
                'ns2:checkVatResponse'
            )
            valid_node = get_first_child_element(
                check_vat_response_node,
                'ns2:valid'
            )
        except Exception as e:
            result.log_lines.append(u'< Response is nondeterministic due to '
                                    u'invalid response body: %r' % (e))
            return result

        # Parse the validity of the business.
        valid_text = get_text(valid_node)

        if valid_text in frozenset(('true', 'false')):
            result.is_valid = valid_text == 'true'
        else:
            result.log_lines.append(u'< Response is nondeterministic due to '
                                    u'invalid validity field: %r' %
                                    (valid_text))

        # Parse the business name and address if possible.
        try:
            name_node = get_first_child_element(
                check_vat_response_node,
                'ns2:name'
            )
            result.business_name = get_text(name_node).strip() or None
        except Exception:
            pass

        try:
            address_node = get_first_child_element(
                check_vat_response_node,
                'ns2:address'
            )
            result.business_address = get_text(address_node).strip() or None
        except Exception:
            pass

        return result


__all__ = ('Registry', 'ViesRegistry', )



from enum import Enum
from .utils import ensure_decimal


class VatChargeAction(Enum):
    """VAT charge action.
    """

    charge = 1
    """Charge VAT.
    """

    reverse_charge = 2
    """No VAT charged but customer is required to account via reverse-charge.
    """

    no_charge = 3
    """No VAT charged.
    """


class VatCharge(object):
    """VAT charge.

    :ivar action: VAT charge action.
    :type action: VatChargeAction
    :ivar country_code:
        Country in which the action applies as the ISO 3166-1 alpha-2 code of
        the country.
    :type country_code: str
    :ivar rate:
        VAT rate in percent. I.e. a value of 25 indicates a VAT charge of 25 %.
    :type rate: Decimal
    """

    def __init__(self, action, country_code, rate):
        self.action = action
        self.country_code = country_code
        self.rate = ensure_decimal(rate)

    def __repr__(self):
        return '<%s.%s: action = %r, country code = %r, rate = %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.action,
            self.country_code,
            self.rate,
        )



class VatNumberCheckResult(object):
    """Result of a VAT number validation check.

    :ivar is_valid:
        Boolean value indicating if the checked VAT number was deemed to be
        valid. ``True`` if the VAT number is valid or ``False`` if the VAT
        number is positively invalid.
    :ivar log_lines:
        Check log lines.
    :ivar business_name: Optional business name retrieved for the VAT number.
    :ivar business_address: Optional address retrieved for the VAT number.
    """

    def __init__(self,
                 is_valid=None,
                 log_lines=None,
                 business_name=None,
                 business_address=None):
        self.is_valid = is_valid
        self.log_lines = log_lines or []
        self.business_name = business_name
        self.business_address = business_address



class NodeNotFoundError(Exception):
    """XML node was not found.
    """

    pass


def get_first_child_element(node, tag_name):
    """Get the first child element node with a given tag name.

    :param node: Parent node.
    :type node: xml.dom.Node
    :returns: the first child element node with the given tag name.
    :rtype: xml.dom.Node
    :raises NodeNotFoundError:
        if no child node with the given tag name was found.
    """

    for child in node.childNodes:
        if child.nodeType == node.ELEMENT_NODE and \
           child.tagName == tag_name:
            return child

    raise NodeNotFoundError('no child element node with tag %s was found' %
                            (tag_name))


def get_text(node):
    """Get text from a node.

    :param node: Node.
    :type node: xml.dom.Node
    :returns: the node's text
    """

    return ''.join(child.data for child in node.childNodes
                   if child.nodeType == node.TEXT_NODE)



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

packages = [
    'pyvat',
]

requires = [
    'requests>=1.0.0,<3.0',
    'pycountry',
    'enum34; python_version < "3.4"',
]

tests_require = [
    'nose',
    'rednose',
    'flake8',
    'unittest2',
]

setup(
    name='pyvat',
    version='1.3.16',
    description='VAT validation for Python',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author='Iconfinder',
    author_email='support@iconfinder.com',
    url='https://www.iconfinder.com',
    project_urls={
        'Issue Tracker': 'https://github.com/iconfinder/pyvat/issues',
    },
    packages=packages,
    install_requires=requires,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.7',
    ),
)



from enum import Enum


class ItemType(Enum):
    """Item type.

    If no item type matches the type of item being sold, it is currently not
    supported by pyvat.
    """

    generic_physical_good = 1
    """Generic physical good.
    """

    generic_electronic_service = 2
    """Generic electronic service.

    Any electronic service that is not covered by a more specific item type.
    """

    generic_telecommunications_service = 3
    """Generic telecommunications service.

    Any telecommunications service that is not covered by a more specific item
    type.
    """

    generic_broadcasting_service = 4
    """Generic broadcasting service.

    Any broadcasting service that is not covered by a more specific item type.
    """

    prepaid_broadcasting_service = 5
    """Pre-paid service provided by broadcasting company.
    """

    ebook = 6
    """E-book.
    """

    enewspaper = 7
    """E-newspaper.
    """

    @property
    def is_electronic_service(self):
        return self in frozenset([ItemType.generic_electronic_service,
                                  ItemType.ebook,
                                  ItemType.enewspaper])

    @property
    def is_telecommunications_service(self):
        return self in frozenset([ItemType.generic_telecommunications_service])

    @property
    def is_broadcasting_service(self):
        return self in frozenset([ItemType.generic_broadcasting_service,
                                  ItemType.prepaid_broadcasting_service])



import re
import pycountry
from .item_type import ItemType
from .party import Party
from .registries import ViesRegistry
from .result import VatNumberCheckResult
from .vat_charge import VatCharge, VatChargeAction
from .vat_rules import VAT_RULES


__version__ = '1.3.16'


WHITESPACE_EXPRESSION = re.compile(r'[\s\-]+')
"""Whitespace expression.

Used for cleaning VAT numbers.
"""


VAT_NUMBER_EXPRESSIONS = {
    'AT': re.compile(r'^U\d{8}$', re.IGNORECASE),
    'BE': re.compile(r'^\d{9,10}$'),
    'BG': re.compile(r'^\d{9,10}$'),
    'CY': re.compile(r'^\d{8}[a-z]$', re.IGNORECASE),
    'CZ': re.compile(r'^\d{8,10}$'),
    'DE': re.compile(r'^\d{9}$'),
    'DK': re.compile(r'^\d{8}$'),
    'EE': re.compile(r'^\d{9}$'),
    'ES': re.compile(r'^[\da-z]\d{7}[\da-z]$', re.IGNORECASE),
    'FI': re.compile(r'^\d{8}$'),
    'FR': re.compile(r'^[\da-hj-np-z]{2}\d{9}$', re.IGNORECASE),
    'GR': re.compile(r'^\d{9}$'),
    'HR': re.compile(r'^\d{11}$'),
    'HU': re.compile(r'^\d{8}$'),
    'IE': re.compile(r'^((\d{7}[a-z])|(\d[a-z]\d{5}[a-z])|(\d{6,7}[a-z]{2}))$',
                     re.IGNORECASE),
    'IT': re.compile(r'^\d{11}$'),
    'LT': re.compile(r'^((\d{9})|(\d{12}))$'),
    'LU': re.compile(r'^\d{8}$'),
    'LV': re.compile(r'^\d{11}$'),
    'MT': re.compile(r'^\d{8}$'),
    'NL': re.compile(r'^\d{9}B\d{2}$', re.IGNORECASE),
    'PL': re.compile(r'^\d{10}$'),
    'PT': re.compile(r'^\d{9}$'),
    'RO': re.compile(r'^\d{2,10}$'),
    'SE': re.compile(r'^\d{12}$'),
    'SI': re.compile(r'^\d{8}$'),
    'SK': re.compile(r'^\d{10}$'),
}
"""VAT number expressions.

Mapping form ISO 3166-1-alpha-2 country codes to the whitespace-less expression
a valid VAT number from the given country must match excluding the country code
prefix.

EU VAT number structures are retrieved from `VIES
<http://ec.europa.eu/taxation_customs/vies/faqvies.do>`_.
"""


VIES_REGISTRY = ViesRegistry()
"""VIES registry instance.
"""


VAT_REGISTRIES = {
    'AT': VIES_REGISTRY,
    'BE': VIES_REGISTRY,
    'BG': VIES_REGISTRY,
    'CY': VIES_REGISTRY,
    'CZ': VIES_REGISTRY,
    'DE': VIES_REGISTRY,
    'DK': VIES_REGISTRY,
    'EE': VIES_REGISTRY,
    'ES': VIES_REGISTRY,
    'FI': VIES_REGISTRY,
    'FR': VIES_REGISTRY,
    'GR': VIES_REGISTRY,
    'HU': VIES_REGISTRY,
    'HR': VIES_REGISTRY,
    'IE': VIES_REGISTRY,
    'IT': VIES_REGISTRY,
    'LT': VIES_REGISTRY,
    'LU': VIES_REGISTRY,
    'LV': VIES_REGISTRY,
    'MT': VIES_REGISTRY,
    'NL': VIES_REGISTRY,
    'PL': VIES_REGISTRY,
    'PT': VIES_REGISTRY,
    'RO': VIES_REGISTRY,
    'SE': VIES_REGISTRY,
    'SK': VIES_REGISTRY,
    'SI': VIES_REGISTRY,
}
"""VAT registries.

Mapping from ISO 3166-1-alpha-2 country codes to the VAT registry capable of
validating the VAT number.
"""


def decompose_vat_number(vat_number, country_code=None):
    """Decompose a VAT number and an optional country code.

    :param vat_number: VAT number.
    :param country_code:
        Optional country code. Default ``None`` prompting detection from the
        VAT number.
    :returns:
        a :class:`tuple` containing the VAT number and country code or
        ``(vat_number, None)`` if decomposition failed.
    """

    # Clean the VAT number.
    vat_number = WHITESPACE_EXPRESSION.sub('', vat_number).upper()

    # Attempt to determine the country code of the VAT number if possible.
    if not country_code:
        country_code = vat_number[0:2]

        if any(c.isdigit() for c in country_code):
            # Country code should not contain digits
            return (vat_number, None)

        # Non-ISO code used for Greece.
        if country_code == 'EL':
            country_code = 'GR'

        if country_code not in VAT_REGISTRIES:
            try:
                if not pycountry.countries.get(alpha_2=country_code):
                    return (vat_number, None)
            except KeyError:
                # country code not found
                return (vat_number, None)
        vat_number = vat_number[2:]
    elif vat_number[0:2] == country_code:
        vat_number = vat_number[2:]
    elif country_code == 'GR' and vat_number[0:2] == 'EL':
        vat_number = vat_number[2:]

    return vat_number, country_code


def is_vat_number_format_valid(vat_number, country_code=None):
    """Test if the format of a VAT number is valid.

    :param vat_number: VAT number to validate.
    :param country_code:
        Optional country code. Should be supplied if known, as there is no
        guarantee that naively entered VAT numbers contain the correct alpha-2
        country code prefix for EU countries just as not all non-EU countries
        have a reliable country code prefix. Default ``None`` prompting
        detection.
    :returns:
        ``True`` if the format of the VAT number can be fully asserted as valid
        or ``False`` if not.
    """

    vat_number, country_code = decompose_vat_number(vat_number, country_code)

    if not vat_number or not country_code:
        return False
    elif not any(c.isdigit() for c in vat_number):
        return False
    elif country_code not in VAT_NUMBER_EXPRESSIONS:
        return False
    elif not VAT_NUMBER_EXPRESSIONS[country_code].match(vat_number):
        return False

    return True


def check_vat_number(vat_number, country_code=None):
    """Check if a VAT number is valid.

    If possible, the VAT number will be checked against available registries.

    :param vat_number: VAT number to validate.
    :param country_code:
        Optional country code. Should be supplied if known, as there is no
        guarantee that naively entered VAT numbers contain the correct alpha-2
        country code prefix for EU countries just as not all non-EU countries
        have a reliable country code prefix. Default ``None`` prompting
        detection.
    :returns:
        a :class:`VatNumberCheckResult` instance containing the result for
        the full VAT number check.
    """

    # Decompose the VAT number.
    vat_number, country_code = decompose_vat_number(vat_number, country_code)
    if not vat_number or not country_code:
        return VatNumberCheckResult(False, [
            '> Unable to decompose VAT number, resulted in %r and %r' %
            (vat_number, country_code)
        ])

    # Test the VAT number format.
    format_result = is_vat_number_format_valid(vat_number, country_code)
    if format_result is not True:
        return VatNumberCheckResult(format_result, [
            '> VAT number validation failed: %r' % (format_result)
        ])

    # Attempt to check the VAT number against a registry.
    if country_code not in VAT_REGISTRIES:
        return VatNumberCheckResult()

    return VAT_REGISTRIES[country_code].check_vat_number(vat_number,
                                                         country_code)


def get_sale_vat_charge(date,
                        item_type,
                        buyer,
                        seller):
    """Get the VAT charge for performing the sale of an item.

    Currently only supports determination of the VAT charge for
    telecommunications, broadcasting and electronic services in the EU.

    :param date: Sale date.
    :type date: datetime.date
    :param item_type: Type of the item being sold.
    :type item_type: ItemType
    :param buyer: Buyer.
    :type buyer: Party
    :param seller: Seller.
    :type seller: Party
    :rtype: VatCharge
    """

    # Only telecommunications, broadcasting and electronic services are
    # currently supported.
    if not item_type.is_electronic_service and \
       not item_type.is_telecommunications_service and \
       not item_type.is_broadcasting_service:
        raise NotImplementedError(
            'VAT charge determination for items that are not '
            'telecommunications, broadcasting or electronic services is '
            'currently not supported'
        )

    # Determine the rules for the countries in which the buyer and seller
    # reside.
    buyer_vat_rules = VAT_RULES.get(buyer.country_code, None)
    seller_vat_rules = VAT_RULES.get(seller.country_code, None)

    # Test if the country to which the item is being sold enforces specific
    # VAT rules for selling to the given country.
    if buyer_vat_rules:
        try:
            return buyer_vat_rules.get_sale_to_country_vat_charge(date,
                                                                  item_type,
                                                                  buyer,
                                                                  seller)
        except NotImplementedError:
            pass

    # Fall back to applying VAT rules for selling from the seller's country.
    if seller_vat_rules:
        try:
            return seller_vat_rules.get_sale_from_country_vat_charge(date,
                                                                     item_type,
                                                                     buyer,
                                                                     seller)
        except NotImplementedError:
            pass

    # Nothing we can do from here.
    raise NotImplementedError(
        'cannot determine VAT charge for a sale of item %r between %r and %r' %
        (item_type, seller, buyer)
    )


__all__ = (
    'check_vat_number',
    'get_sale_vat_charge',
    'is_vat_number_format_valid',
    ItemType.__name__,
    Party.__name__,
    VatCharge.__name__,
    VatChargeAction.__name__,
)
